
__fields_info__ = {  "HII_fraction"                          : { "synonyms" : ['ionized_fraction', 'x_HII', 'f_HII', 'x_ion', 'f_ion', 'ion_frac']                                     , "description" : "Hydrogen ionized fraction"                                   , "symbol" : r"$x_\mathrm{HII}$"                                                          , "units" : ""             , "remarks": "volume-averaged", "required_fields" : ["redshift"]}  ,
                     "HeIII_fraction"                        : { "synonyms" : ['helium_ionized_fraction', 'ionized_fraction_helium', 'He_ion_frac', 'ion_frac_He']                     , "description" : "Helium doubly-ionized fraction"                              , "symbol" : r"$x_\mathrm{HeIII}$"                                                        , "units" : ""             , "remarks": "volume-averaged", "required_fields" : ["redshift"]}  ,
                     "Lya_flux_power_spectrum"               : { "synonyms" : ['Lya_flux_PS']                                                                                          , "description" : "Ly-alpha flux power spectrum"                                , "symbol" : r"$P\, [\mathrm{km/s}]$"                                                     , "units" : "km/s"         , "remarks": ""               , "required_fields" : ["redshift"]}  , 
                     "mean_free_path"                        : { "synonyms" : ['mfp', 'lambda_mfp', 'lambda_ion', 'lambda_912', 'mfp_912', 'mfp_ion']                                  , "description" : "Mean free path of ionising photons"                          , "symbol" : r"$\lambda_\mathrm{mfp} \, [h^{-1} \, \mathrm{Mpc}]$"                        , "units" : "Mpc/h"        , "remarks": ""               , "required_fields" : ["redshift"]}  , 
                     "effective_optical_depth_HI_Lya"        : { "synonyms" : ['tau_HI', 'tau_eff_HI']                                                                                 , "description" : "HI Ly-alpha effective optical depth"                         , "symbol" : r"$\tau_\mathrm{eff, HI}$"                                                   , "units" : ""             , "remarks": ""               , "required_fields" : ["redshift"]}  , 
                     "effective_optical_depth_HeII_Lya"      : { "synonyms" : ['tau_HeII', 'tau_eff_HeII']                                                                             , "description" : "HeII Ly-alpha effective optical depth"                       , "symbol" : r"$\tau_\mathrm{eff, HeII}$"                                                 , "units" : ""             , "remarks": ""               , "required_fields" : ["redshift"]}  , 
                     "HeII_to_HI_column_density_ratio"       : { "synonyms" : ['eta']                                                                                                  , "description" : "HeII-to-HI column density ratio"                             , "symbol" : r"$\eta$"                                                                    , "units" : ""             , "remarks": ""               , "required_fields" : ["redshift"]}  , 
                     "quasar_luminosity_function"            : { "synonyms" : ['QLF', 'QSOLF']                                                                                         , "description" : "Quasar luminosity function"                                  , "symbol" : r"$\log_{10} (phi_\mathrm{QSO}) \, [\mathrm{Mpc}^{-3} \, \mathrm{mag}^{-1}]$", "units" : "Mpc^-3 mag^-1", "remarks": ""               , "required_fields" : ["redshift", "M_1450"]}  , 
                     "UV_luminosity_function"                : { "synonyms" : ['UVLF']                                                                                                 , "description" : "Galaxy UV luminosity function"                               , "symbol" : r"$\log_{10} (phi_\mathrm{UV}) \, [\mathrm{Mpc}^{-3} \, \mathrm{mag}^{-1}]$" , "units" : "Mpc^-3 mag^-1", "remarks": ""               , "required_fields" : ["redshift", "M_UV"]  }  , 
                     "IGM_temperature_mean_density"          : { "synonyms" : ['T0']                                                                                                   , "description" : "IGM temperature at mean density"                             , "symbol" : r"T0 [K]"                                                                    , "units" : "K"            , "remarks": ""               , "required_fields" : ["redshift"]}  , 
                     "optical_depth_CMB"                     : { "synonyms" : ['tau_cmb', 'tau_CMB']                                                                                   , "description" : "CMB optical depth"                                           , "symbol" : r"$\tau_\mathrm{CMB}$"                                                       , "units" : ""             , "remarks": ""               , "required_fields" : []          }  , 
                     "sfrd"                                  : { "synonyms" : ['SFRD', 'star formation rate density', 'star-formation-rate density', 'star-formation rate density']    , "description" : "Star-formation-rate density"                                 , "symbol" : r"$\log \Psi \, [\mathrm{M}_\odot \mathrm{yr}^{-1} \mathrm{Mpc}^{-3}]$"      , "units" : "Msun/yr/Mpc^3", "remarks": ""               , "required_fields" : ["redshift"]}  , 
                     "Lya_spike_galaxy_correlation"          : { "synonyms" : []                                                                                                       , "description" : "Ly-alpha spike - galaxy correlation"                         , "symbol" : r"$\langle f \rangle / \bar{f} - 1$"                                         , "units" : ""             , "remarks": ""               , "required_fields" : ["redshift"]}  ,
                     "mass_stellar_metallicity_relation"     : { "synonyms" : ['stellar MZR']                                                                                          , "description" : "Stellar metallicity as function of galaxy stellar mass"      , "symbol" : r"$\log(Z/Z_\odot)$"                                                         , "units" : ""             , "remarks": ""               , "required_fields" : ["redshift", "stellar mass"]}  ,
                     "mass_gas_metallicity_relation"         : { "synonyms" : ['gas MZR']                                                                                              , "description" : "Gas-phase O/H metallicity as function of galaxy stellar mass", "symbol" : r"$\log(O/H) + 12$"                                                          , "units" : ""             , "remarks": ""               , "required_fields" : ["redshift", "stellar mass"]}  ,
                     "galaxy_main_sequence"                  : { "synonyms" : ['GMS']                                                                                                  , "description" : "Star formation rate  as function of galaxy stellar mass"     , "symbol" : r"SFR [M$_\odot$/yr]"                                                        , "units" : "Msun/yr"      , "remarks": ""               , "required_fields" : ["redshift", "stellar mass"]}  ,
                     "UV_slope"                              : { "synonyms" : ['beta_UV']                                                                                              , "description" : "Slope of the galaxy UV log flux density"                     , "symbol" : r"$\beta$"                                                                   , "units" : ""             , "remarks": ""               , "required_fields" : ["redshift"]}  ,
                     "ionizing_photons_production_efficiency": { "synonyms" : ['xi_ion', 'csi_ion', 'ionizing_photon_production_efficiency', 'UV_to_ionizing_photons_conversion']      , "description" : "Number of ionizing photons produced per unit UV luminosity"  , "symbol" : r"$\log(\xi_\mathrm{ion} [\mathrm{Hz\,erg^{-1}}])$"                          , "units" : "Hz erg^-1"    , "remarks": ""               , "required_fields" : ["redshift"]}  ,
                     "HI_photoionization_rate"               : { "synonyms" : ['photoionization_rate', 'photoionisation_rate', 'Gamma', 'GammaHI']                                                             , "description" : "Photoionsation rate of hydrogen in the IGM"                  , "symbol" : r"$\Gamma_\mathrm{HI}$ [s$^{-1}$]"                                           , "units" : "s^-1"         , "remarks": ""               , "required_fields" : ["redshift"]}  ,
                     "ionizing_photons_emission_rate"        : { "synonyms" : ['ionizing_photons_production_rate', 'dotN', 'dotNion', 'Nion', 'N_ion']                                 , "description" : "Emission rate of ionizing photons per unit volume"           , "symbol" : r"$\dot{N}_\mathrm{ion}$ [s$^{-1}$ Mpc$^{-1}$]"                              , "units" : "s^-1 Mpc^-3"  , "remarks": ""               , "required_fields" : ["redshift"]}  ,
                     "reionization_midpoint"                 : { "synonyms" : ['z_mid']                                                                                                , "description" : "Redshift when x_HI = 0.5"                                    , "symbol" : r"$z_\mathrm{mid}$"                                                          , "units" : ""             , "remarks": ""               , "required_fields" : []          }  ,
                     "UV_luminosity_density"                 : { "synonyms" : ['rho_UV']                                                                                               , "description" : "Density of UV radiation in erg/s/Hz/Mpc^3"                   , "symbol" : r"$\log_{10} (\rho_\mathrm{UV}) \, [\mathrm{erg} \, \mathrm{s}^{-1} \, \mathrm{Hz}^{-1} \, \mathrm{Mpc}^{-3})$" , "units" : "erg s^-1 Hz^-1 Mpc^-3", "remarks": "", "required_fields" : ["redshift"]  }  , 
                  }
